/**
 ******************************************************************************
 * @file           : main.c
 * @author         : Auto-generated by STM32CubeIDE
 * @brief          : Main program body
 ******************************************************************************
 * @attention
 *
 * Copyright (c) 2022 STMicroelectronics.
 * All rights reserved.
 *
 * This software is licensed under terms that can be found in the LICENSE file
 * in the root directory of this software component.
 * If no LICENSE file comes with this software, it is provided AS-IS.
 *
 ******************************************************************************
 */

#include <stdint.h>

#if !defined(__SOFT_FP__) && defined(__ARM_FP)
  #warning "FPU is not initialized, but the project is compiling for an FPU. Please initialize the FPU before use."
#endif

#define  SET_UNPRIV_LEVEL   __asm volatile(                 \
                                "MRS R0, CONTROL\n\t"       \
                                "ORR R0, R0, #1\n\t"        \
                                "MSR CONTROL, r0"           \
                            );

#define  SET_PRIV_LEVEL     __asm volatile(                 \
                                "MRS R0, CONTROL\n\t"       \
                                "BIC R0, R0, #1\n\t"        \
                                "MSR CONTROL, r0"           \
                            );

void SVC_Handler(void)
{
  SET_PRIV_LEVEL;
}

void myfunc() {
  __asm volatile("NOP");
}

int main(void)
{
  /* Basic Assembly */
  __asm volatile(
    "LDR R1, =#0x20001000\n\t" /* Load address 0x20001000 to R1 */
    "LDR R2, =#0x20001004\n\t" /* Load address 0x20001004 to R2 */
    "LDR R3, =#0x20001008\n\t" /* Load address 0x20001008 to R3 */
    "LDR R0, [R1]\n\t"         /* Load data at the address pointing by R1, save to R0 */
    "LDR R1, [R2]\n\t"         /* Load data at the address pointing by R2, save to R1 */
    "ADD R0, R1\n\t"           /* Add R0 to R1, save to R0 */
    "STR R0, [R3]\n\t"         /* Store R0 to the address pointing by R3 */
  );

  /* Advanced Assembly */
  int val = 50;
  __asm volatile("MOV R0, %0": : "r"(val)); // R0 = 50
  __asm volatile("MOV R1, %0": : "i"(50));  // R1 = 50

  int control_reg;
  __asm volatile("MRS %0, CONTROL": "=r"(control_reg)); // control_reg = CONTROL

  int var1=10;
  int var2;
  __asm volatile("MOV %0, %1": "=r"(var2): "r"(var1));  // var2 = var1

  int p1, *p2;
  p2 = (int*) 0x20000000;
  __asm volatile("LDR %0, [%1]": "=r"(p1) : "r"(p2)); // pi = *p2

  register int *foo __asm ("r12"); // foo is R12 register
  *foo = 12;

  /* Change to Unprivileged level */
  SET_UNPRIV_LEVEL;

  void (*pfunc_by_name)() = myfunc; // at 0x80001f2, but compiler assign 0x80001f3
  pfunc_by_name();

  // to look up the address of myfunc in Debug/*.list file
  // however, the address is not fixed, it may change in every build

#if 0
  // using the correct address of a function leads to Usage Fault Exception
  void (*pfunc_by_addr)() = (void *)0x080001f2;
#else
  // using the modified address of a function by adding it with 1 will work normally
  void (*pfunc_by_addr)() = (void *)0x080001f3;
#endif
  pfunc_by_addr();

  /* Back to Privileged level through SVC interrupt*/
  __asm volatile("SVC #0");

  /* Loop forever */
	for(;;);
}
